-- Copyright (c) 2009, ERICSSON AB
-- All rights reserved.
--
-- Redistribution and use in source and binary forms, with or without
-- modification, are permitted provided that the following conditions are met:
--
--     * Redistributions of source code must retain the above copyright notice,
--       this list of conditions and the following disclaimer.
--     * Redistributions in binary form must reproduce the above copyright
--       notice, this list of conditions and the following disclaimer in the
--       documentation and/or other materials provided with the distribution.
--     * Neither the name of the ERICSSON AB nor the names of its contributors
--       may be used to endorse or promote products derived from this software
--       without specific prior written permission.
--
-- THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
-- AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-- IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
-- DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
-- FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
-- DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
-- SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
-- CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
-- OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
-- OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

{-# LANGUAGE FlexibleInstances, GADTs, TypeFamilies, TypeSynonymInstances #-}



--------------------------------------------------------------------------------
-- This module defines the type representing Feldspar expressions
-- (`Expr`/`Data`).
--------------------------------------------------------------------------------

module Feldspar.Core.Expr where



import Feldspar.Core.Ref



type Length = Int

-- Typed core language expressions. A value of type `Expr a` is a representation
-- of a program that computes a value of type `a`.
--
-- The semantics of `Expr` is given by `evalE`.
data Expr a
  where
    Input      :: Expr a
    Value      :: Show a => a -> Expr a
    Pair       :: Data a -> Data b -> Expr (a,b)
    Function   :: String -> (a -> b) -> (Data a -> Expr b)
    IfThenElse :: Data Bool -> Data a -> Data a -> Expr a
    While   :: (Data a -> Data Bool) -> (Data a -> Data a) -> (Data a -> Expr a)
    Parallel   :: Data Length -> (Data Int -> Data a) -> Expr [a]



-- A wrapper around `Expr` to allow observable sharing (see module
-- Feldspar.Core.Ref).
type Data a = Ref (Expr a)



-- The semantics of expressions
evalE :: Expr a -> a

evalE Input              = error "evaluating Input"
evalE (Value a)          = a
evalE (Pair a b)         = (evalD a, evalD b)
evalE (Function _ f a)   = f (evalD a)
evalE (IfThenElse c t e) = if evalD c then evalD t else evalD e

evalE (While continue body init) = loop init
  where
    loop s = if done then evalD s else loop (body s)
      where
        done = not $ evalD $ continue s

evalE (Parallel l ixf) = map (evalD . ixf . value) [0 .. n-1]
  where
    n = evalD l



-- The semantics of `Data`
evalD :: Data a -> a
evalD = evalE . deref



-- Computable types. A computable value completely represents a core program, in
-- such a way that `internalize . externalize` preserves semantics, but not
-- necessarily syntax.
--
-- The terminology used in this class comes from thinking of the `Data` type as
-- the "internal" core language and the Feldspar.Core API as the "external" core
-- language.
class Computable a
  where
    -- `Data (Internal a)` is the internal representation of the type `a`.
    type Internal a

    -- Convert to internal representation
    internalize :: a -> Data (Internal a)

    -- Convert to external representation
    externalize :: Data (Internal a) -> a

instance Computable (Data a)
  where
    type Internal (Data a) = a

    internalize = id
    externalize = id

instance (Computable a, Computable b) => Computable (a,b)
  where
    type Internal (a,b) = (Internal a, Internal b)

    internalize (a,b) = ref $ Pair (internalize a) (internalize b)
    externalize ab    = (externalize (first ab), externalize (second ab))

-- Lower a function to operate on internal representation.
lowerFun :: (Computable a, Computable b) =>
    (a -> b) -> (Data (Internal a) -> Data (Internal b))
lowerFun f = internalize . f . externalize

-- Lift a function to operate on external representation.
liftFun :: (Computable a, Computable b) =>
    (Data (Internal a) -> Data (Internal b)) -> (a -> b)
liftFun f = externalize . f . internalize



-- The semantics of any `Computable` type
eval :: Computable a => a -> Internal a
eval = evalD . internalize



-- A program that computes a constant value
value :: Show a => a -> Data a
value = ref . Value

unit :: Data ()
unit = value ()

true :: Data Bool
true = value True

false :: Data Bool
false = value False



-- Constructs a one-argument primitive function.
--
-- function fun f:
--
--   * `fun` is the name of the function.
--   * `f` gives the evaluation semantics.
function :: Show b => String -> (a -> b) -> (Data a -> Data b)
function fun f a = case deref a of
    Value a' -> ref $ Value $ f a'
    _        -> ref $ Function fun f a

-- A two-argument primitive function
function2 :: Show c => String -> (a -> b -> c) -> (Data a -> Data b -> Data c)
function2 fun f a b = case (deref a, deref b) of
    (Value a', Value b') -> ref $ Value $ f a' b'
    _                    -> ref $ Function fun f' $ ref $ Pair a b
  where
    f' (a,b) = f a b

-- A three-argument primitive function
function3 :: Show d =>
    String -> (a -> b -> c -> d) -> (Data a -> Data b -> Data c -> Data d)
function3 fun f a b c = case (deref a, deref b, deref c) of
    (Value a', Value b', Value c') -> ref $ Value $ f a' b' c'
    _ -> ref $ Function fun f' $ ref $ Pair a $ ref $ Pair b c
  where
    f' (a,(b,c)) = f a b c



instance Show (Data a)
  where
    show a = "... :: Data a"
  -- Needed for the `Num` instance.

instance Num a => Num (Data a)
  where
    fromInteger = value . fromInteger
    abs         = function  "abs"    abs
    signum      = function  "signum" signum
    (+)         = function2 "(+)"    (+)
    (-)         = function2 "(-)"    (-)
    (*)         = function2 "(*)"    (*)

instance Fractional (Data Float)
  where
    fromRational = value . fromRational
    (/)          = function2 "(/)" (/)



first :: Data (a,b) -> Data a
first = ref . Function "first" fst

second :: Data (a,b) -> Data b
second = ref . Function "second" snd

-- Look up an index in an array (see also `!`)
getIx :: Show a => Data [a] -> Data Int -> Data a
getIx arr = function2 "(!)" f arr
  where
    f as i
        | i >= la || i < 0 = error "getIx: index out of bounds"
        | otherwise        = as !! i
      where
        la   = length as

-- setIx arr i a:
--
-- Replaces the value at index `i` in the array `arr` with the value `a`.
setIx :: Show a => Data [a] -> Data Int -> Data a -> Data [a]
setIx arr = function3 "setIx" f arr
  where
    f as i a
        | i >= la || i < 0 = error "getIx: index out of bounds"
        | otherwise        = take i as ++ [a] ++ drop (i+1) as
      where
        la = length as

class RandomAccess a
  where
    -- The type of elements in a random access structure
    type Element a

    -- Index lookup in a random access structure
    (!) :: a -> Data Int -> Element a

instance Show a => RandomAccess (Data [a])
  where
    type Element (Data [a]) = Data a
    (!) = getIx

-- ifThenElse cond t e:
--
-- Selects between `t` and `e` depending on whether the condition `cond` is true
-- or false.
ifThenElse :: Computable a => Data Bool -> a -> a -> a
ifThenElse cond t e = case deref cond of
    Value True  -> t
    Value False -> e
    _ -> externalize $ ref $ IfThenElse cond (internalize t) (internalize e)



-- A while-loop
--
-- while cont body :: state -> state:
--
--   * `state` is the type of the state.
--   * `cont` determines whether or not to continue based on the current state.
--   * `body` computes the next state from the current state.
--   * The result is a function from initial state to final state.
while
    :: Computable state
    => (state -> Data Bool)
    -> (state -> state)
    -> (state -> state)

while cont body = liftFun (ref . While (lowerFun cont) (lowerFun body))



-- A parallel array
--
-- parallel l ixf:
--
--   * `l` is the length of the resulting array (outermost level).
--   * `ifx` is a function that maps each index in the range `[0 .. l-1]` to its
--     element.
--
-- Since there are no dependencies between the elements, the compiler is free to
-- compute the elements in any order, or even in parallel.
parallel :: Data Length -> (Data Int -> Data a) -> Data [a]
parallel l ixf = ref $ Parallel l ixf

