-- Copyright (c) 2009, ERICSSON AB
-- All rights reserved.
--
-- Redistribution and use in source and binary forms, with or without
-- modification, are permitted provided that the following conditions are met:
--
--     * Redistributions of source code must retain the above copyright notice,
--       this list of conditions and the following disclaimer.
--     * Redistributions in binary form must reproduce the above copyright
--       notice, this list of conditions and the following disclaimer in the
--       documentation and/or other materials provided with the distribution.
--     * Neither the name of the ERICSSON AB nor the names of its contributors
--       may be used to endorse or promote products derived from this software
--       without specific prior written permission.
--
-- THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
-- AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-- IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
-- DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
-- FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
-- DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
-- SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
-- CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
-- OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
-- OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

{-# LANGUAGE FlexibleInstances, OverlappingInstances, ScopedTypeVariables, UndecidableInstances #-}



--------------------------------------------------------------------------------
-- Functions for reifying expressions to textual format.
--------------------------------------------------------------------------------

module Feldspar.Core.Reify
    ( Program (..)
    , printCore
    ) where



import Control.Monad.State
import Control.Monad.Writer
import Data.List
import Data.Map (Map)
import qualified Data.Map as Map
import Data.Maybe
import Data.Unique

import Feldspar.Core.Ref
import Feldspar.Core.Expr



type NodeId = Int

type Node = String

data Info = Info
  { index   :: NodeId             -- Next id
  , visited :: Map Unique NodeId  -- Visited references mapped to their id
  }

-- Monad for making graph building easier
type Reify a = WriterT [Node] (State Info) a

startInfo :: Info
startInfo = Info 0 Map.empty

runGraph :: Reify a -> Info -> (a, [Node])
runGraph graph info = evalState (runWriterT graph) info

newIndex :: Reify NodeId
newIndex = do
    info <- get
    put (info {index = succ (index info)})
    return (index info)

remember :: Data a -> NodeId -> Reify ()
remember a i = modify $ \info ->
    info {visited = Map.insert (refId a) i (visited info)}

checkNode :: Data a -> Reify (Maybe NodeId)
checkNode a = gets ((Map.lookup (refId a)) . visited)

var :: NodeId -> String
var i = "v" ++ show i

lambda :: (NodeId, NodeId) -> String
lambda (inp,outp) = "(\\" ++ var inp ++ " -> " ++ var outp ++ ")"

-- Declare a node
node :: Data a -> String -> [NodeId] -> Reify NodeId
node a str inps = do
    i <- newIndex
    remember a i
    tell [var i ++ " = " ++ str ++ inps']
    return i
  where
    inps' = case inps of
      []    -> ""
      [inp] -> var inp
      _     -> "(" ++ intercalate "," (map var inps) ++ ")"



buildGraph :: Data a -> Reify NodeId
buildGraph a = do
    ia <- checkNode a
    case ia of
      Nothing -> list (deref a)
      Just ia -> return ia
  where
    funcNode fun b = do
      ib <- buildGraph b
      node a (fun ++ " ") [ib]

    list :: Expr a -> Reify NodeId

    list Input     = censor (const []) $ node a "" []
    list (Value b) = node a ("value " ++ show b) []

    list (Pair b c) = do
      ib <- buildGraph b
      ic <- buildGraph c
      node a "" [ib,ic]

    list (Function fun _ b) = funcNode fun b

    list (IfThenElse c t e) = do
      ic <- buildGraph c
      it <- buildGraph t
      ie <- buildGraph e
      node a "ifThenElse " [ic,it,ie]

    list (While cont body b) = do
      contIO <- buildSubFun cont
      bodyIO <- buildSubFun body
      funcNode ("while " ++ lambda contIO ++ " " ++ lambda bodyIO) b

    list (Parallel l ixf) = do
      ixfIO <- buildSubFun ixf
      funcNode ("parallel " ++ lambda ixfIO) l



buildSubFun :: (Data a -> Data b) -> Reify (NodeId, NodeId)
buildSubFun f = do
    let inp  = ref Input
        outp = f inp
    liftM2 (,)
      (buildGraph inp)  -- Needed in case input is not used
      (buildGraph outp)



showCoreD :: Bool -> (Data a -> Data b) -> String
showCoreD hasArg f = unlines $
    [ "program" ++ inStr ++ " = " ++ var outp
    , "  where"
    ]
      ++
    map ("    " ++) nodes
  where
    ((inp,outp),nodes) = runGraph (buildSubFun f) startInfo
    inStr = if hasArg then " " ++ var inp else ""



-- Types that represent core language programs
class Program a
  where
    -- Converts a program to a Graph
    showCore :: a -> String

instance Computable a => Program a
  where
    showCore a = showCoreD False (\(_::Data ()) -> internalize a)

instance (Computable a, Computable b) => Program (a -> b)
  where
    showCore = showCoreD True . lowerFun

instance (Computable a, Computable b, Computable c) => Program (a -> b -> c)
  where
    showCore f = showCoreD True $ lowerFun $ \(a,b) -> f a b

instance (Computable a, Computable b, Computable c, Computable d)
      => Program (a -> b -> c -> d)
  where
    showCore f = showCoreD True $ lowerFun $ \(a,(b,c)) -> f a b c



printCore :: Program a => a -> IO ()
printCore = putStrLn . showCore

